import os

import joblib
import re

from Lutil.checkpoints._check_util import (
    _get_applied_args,
    _get_hash_of_str,
    _get_identify_str_for_func,
    _get_file_info,
    _get_identify_str_for_value,
    _check_handleable,
    _check_inline_handleable,
)

from Lutil._exceptions import SkipWithBlock, InlineEnvironmentWarning
import sys
import inspect
from Lutil._logging import logger
import warnings


_save_dir = ".Lutil-checkpoint"


def checkpoint(ignore=[]):
    if callable(ignore):
        param_is_callable = True
        func = ignore
        ignore = []
    elif isinstance(ignore, (list, tuple)):
        param_is_callable = False
    else:
        raise TypeError(f"Unsupported parameter type '{type(ignore)}'")

    def wrapper(func):
        def inner(*args, **kwargs):
            if not os.path.exists(_save_dir):
                os.mkdir(_save_dir)
            recompute = "__recompute__" in kwargs and kwargs["__recompute__"]
            if recompute:
                kwargs.pop("__recompute__")

            _check_handleable(func)
            file_info = _get_file_info(func)

            applied_args = _get_applied_args(func, args, kwargs)
            id_str = _get_identify_str_for_func(func, applied_args, ignore)
            hash_val = _get_hash_of_str(file_info + id_str)

            cache_path = os.path.join(_save_dir, f"{hash_val}.pkl")

            if os.path.exists(cache_path) and not recompute:
                return joblib.load(cache_path)
            else:
                res = func(*args, **kwargs)
                joblib.dump(res, cache_path)
                return res

        return inner

    if param_is_callable:
        return wrapper(func)
    else:
        return wrapper


class InlineCheckpoint(object):
    def __init__(self, *, watch, produce):
        assert isinstance(watch, (list, tuple))
        assert isinstance(produce, (list, tuple))
        self.watch = watch
        self.produce = produce

        if not os.path.exists(_save_dir):
            os.mkdir(_save_dir)

        call_f = inspect.currentframe().f_back
        self.lineno = call_f.f_lineno
        self.locals = call_f.f_locals
        self.globals = call_f.f_globals

        self.__check_watch_produce()

        status_str = self.__get_status_str()
        self.status_hash = _get_hash_of_str(status_str)

        logger.debug(f"status_str: {status_str}")

        self.skip = self.__check_skip()
        logger.debug(f"skip: {self.skip}")

    def __get_watch(self, i):
        assert isinstance(i, str)
        e = ValueError(f"{i} is not a valid identifier.")
        ref_list = i.split(".")
        if ref_list[0] not in self.locals:
            raise e
        curr = self.locals[ref_list[0]]
        for ref in ref_list[1:]:
            if not hasattr(curr, ref):
                raise e
            else:
                curr = getattr(curr, ref)
        return curr

    def __check_watch_produce(self):
        for i in self.watch:
            self.__get_watch(i)

        for i in self.produce:
            assert isinstance(i, str)
            e = ValueError(f"{i} is not a valid identifier.")
            pattern = "^[a-zA-Z_][a-zA-Z_0-9]*$"

            if "." in i:
                ref_list = i.split(".")
                if ref_list[0] not in self.locals:
                    raise e
                curr = self.locals[ref_list[0]]
                for ref in ref_list[1:-1]:
                    if not hasattr(curr, ref):
                        raise e
                    else:
                        curr = getattr(curr, ref)

                if not re.compile(pattern).match(ref_list[-1]):
                    raise e
            else:
                if not re.compile(pattern).match(i):
                    raise e

    def __get_start_line_and_indent(self, sourcelines):
        pattern = (
            r"""(\s*)with .*?\(\s*watch\s*=\s*[\[\(]\s*['"]?%s['"]?\s*[\]\)]\s*,\s*produce\s*=\s*[\[\(]\s*['"]?%s['"]?\s*[\]\)]\s*\).*?:"""
            % ("""['"]\s*,\s*['"]""".join(self.watch), """['"]\s*,\s*['"]""".join(self.produce))
        )

        matcher = re.compile(pattern)
        start_line = res = None

        initial_lineno = self.lineno

        logger.debug(f"start-line pattern: {pattern}")
        logger.debug(f"initial_lineno: {initial_lineno}")
        logger.debug(f"sourcelines: {sourcelines}")

        while not res and self.lineno > 0:
            init_statement = "\n".join(sourcelines[self.lineno - 1 :])
            res = matcher.match(init_statement)
            self.lineno -= 1

        if not res:
            self.lineno = initial_lineno

        while not res and self.lineno < len(sourcelines):
            init_statement = "\n".join(sourcelines[self.lineno - 1 :])
            res = matcher.match(init_statement)
            self.lineno += 1

        if res:
            start_line = self.lineno
            indent = res.group(1)

        if start_line is None:
            raise Exception("Failed to check the content in the with-statement.")

        return start_line, indent

    def __get_status_str(self):
        watch_dict = {}

        for i in self.watch:
            value = self.__get_watch(i)
            _check_inline_handleable(value)
            if inspect.ismethod(value) or inspect.isfunction(value):
                watch_dict[i] = _get_identify_str_for_func(value)
            else:
                watch_dict[i] = _get_identify_str_for_value(value)

        watch_str = "-".join([f"{k}:{v}" for k, v in watch_dict.items()])

        if "_ih" in self.globals and "In" in self.globals and "__file__" not in self.globals:
            file_name = "jupyter-notebook"
            source = "\n".join(self.globals["In"])
        elif "__file__" in self.globals:
            file_name = os.path.basename(self.globals["__file__"])
            with open(self.globals["__file__"], "r", encoding="utf-8") as f:
                source = f.read()
        else:
            raise Exception("Unknown error when detecting jupyter or .py environment.")

        sourcelines = source.split("\n")
        start_line, indent = self.__get_start_line_and_indent(sourcelines)

        with_statement_lines = []
        for i in range(start_line + 1, len(sourcelines)):
            line = sourcelines[i]
            pattern = indent + r"\s+\S+"
            matcher = re.compile(pattern)
            if matcher.match(line):
                with_statement_lines.append(line)
            else:
                break

        with_statement = ";".join([i.strip() for i in with_statement_lines]).replace(" ", "")

        identify_str = f"{file_name}-{watch_str}-{with_statement}"
        return identify_str

    def __checkpoint_exists(self):
        if not self.produce:
            return os.path.exists(self.__cache_file_name(None))

        for i in self.produce:
            if not os.path.exists(self.__cache_file_name(i)):
                return False
        return True

    def __check_skip(self):
        if "__name__" not in self.locals or self.locals["__name__"] != "__main__":
            for i in self.produce:
                if "." not in i:
                    warnings.warn(InlineEnvironmentWarning())
                    return False

        return self.__checkpoint_exists()

    def __enter__(self):
        if self.skip:
            sys.settrace(lambda *args, **keys: None)
            frame = sys._getframe(1)
            frame.f_trace = self._trace
        return self

    def _trace(self, frame, event, arg):
        raise SkipWithBlock()

    def __exit__(self, type, value, traceback):
        if type is not None and type is not SkipWithBlock:
            return

        if self.skip:
            for i in self.produce:
                self.__retrieve(i)
        elif not self.produce:
            joblib.dump(None, self.__cache_file_name(None))
        else:
            for i in self.produce:
                self.__save(i)

        return True

    def __cache_file_name(self, i):
        if i is None:
            i = "None"

        return os.path.join(_save_dir, f"{self.status_hash}-{i}.pkl")

    def __retrieve(self, i):
        obj = joblib.load(self.__cache_file_name(i))

        if "." not in i:
            self.locals[i] = obj
        else:
            ref_list = i.split(".")
            curr = self.locals[ref_list[0]]
            for ref in ref_list[1:-1]:
                curr = getattr(curr, ref)

            setattr(curr, ref_list[-1], obj)

    def __save(self, i):
        if "." not in i:
            obj = self.locals[i]
        else:
            ref_list = i.split(".")
            curr = self.locals[ref_list[0]]
            for ref in ref_list[1:]:
                curr = getattr(curr, ref)
            obj = curr

        joblib.dump(obj, self.__cache_file_name(i))



import hashlib
import inspect
import os
import re
from collections import OrderedDict

import numpy as np
import pandas as pd
import warnings

from Lutil._exceptions import NotDecoratableError, ComplexParamsIdentifyWarning, NotInlineCheckableError
from Lutil._logging import logger


def _get_file_info(obj):
    filename = inspect.getfile(obj)
    ipython_filename_pattern = r"<ipython-input-\d+-(.{12})>"
    ipython_result = re.compile(ipython_filename_pattern).match(filename)
    if ipython_result:
        file_info = f"ipynb-{ipython_result.group(1)}"
    else:
        file_info, _ = os.path.splitext(os.path.basename(filename))

    return file_info


def _get_identify_str_for_cls_or_object(obj):
    identify_dict = {}
    for attr in dir(obj):
        if attr.startswith("__") and attr.endswith("__"):
            pass
        else:
            value = getattr(obj, attr)
            if (
                inspect.ismethod(value)
                or inspect.isclass(value)
                or inspect.isfunction(value)
                or inspect.isbuiltin(value)
            ):
                pass
            elif isinstance(value, (pd.DataFrame, pd.Series)):
                identify_dict[attr] = _hash_pd_object(value)
            elif isinstance(value, np.ndarray):
                identify_dict[attr] = _hash_np_array(value)
            else:
                str_val = str(value)
                if re.compile(r"<.*? object at \w{12,20}>").match(str_val):
                    warnings.warn(ComplexParamsIdentifyWarning(f"A complicated object is an attribute of {str(obj)}"))
                else:
                    identify_dict[attr] = str_val + str(type(value))

    return "-".join([k + ":" + v for k, v in identify_dict.items()])


def _hash_pd_object(obj):
    try:
        return hashlib.md5(pd.util.hash_pandas_object(obj, index=True).values).hexdigest()
    except TypeError:
        return hashlib.md5(
            pd.util.hash_pandas_object(obj.applymap(_get_identify_str_for_value), index=True).values
        ).hexdigest()


def _hash_np_array(arr):
    return "numpy" + str(type(arr)) + _hash_pd_object(pd.DataFrame(arr))


def _get_identify_str_for_value(value):
    if isinstance(value, (pd.DataFrame, pd.Series)):
        return _hash_pd_object(value)

    elif isinstance(value, np.ndarray):
        return _hash_np_array(value)

    else:
        str_val = str(value)
        if re.compile(r"<.*? object at \w{12,20}>").match(str_val):
            return _get_identify_str_for_cls_or_object(value)
        else:
            return str_val + str(type(value))


def _is_general_handleable(obj):
    return not (
        inspect.isgenerator(obj)
        or inspect.isgeneratorfunction(obj)
        or inspect.iscoroutine(obj)
        or inspect.iscoroutinefunction(obj)
        or inspect.isasyncgen(obj)
        or inspect.isasyncgenfunction(obj)
    )


def _check_handleable(obj):
    if not _is_general_handleable(obj):
        raise NotDecoratableError(obj)


def _check_inline_handleable(obj):
    if not _is_general_handleable(obj):
        raise NotInlineCheckableError(obj)


def _get_identify_str_for_func(func, applied_args, ignore=[]):
    qualname = func.__qualname__

    identify_args = {}
    for key, value in applied_args.items():
        if key in ignore:
            pass

        elif key in ("cls", "self"):
            identify_args[key] = _get_identify_str_for_cls_or_object(value)

        elif inspect.isclass(value):
            warnings.warn(ComplexParamsIdentifyWarning(f"A class is used as the parameter"))
            identify_args[key] = value.__qualname__

        elif inspect.ismethod(value) or inspect.isfunction(value):
            warnings.warn(ComplexParamsIdentifyWarning(f"A function is used as the parameter"))
            tmp_applied_args = _get_applied_args(value, (), {})
            identify_args[key] = _get_identify_str_for_func(value, tmp_applied_args)

        else:
            identify_args[key] = _get_identify_str_for_value(value)

    identify_args_str = "-".join([f"{k}:{v}" for k, v in identify_args.items()])

    code = inspect.getsource(func).replace("\n", "").replace(" ", "")

    full_str = f"{qualname}-{identify_args_str}-{code}"
    logger.debug(f"Identification String: {full_str}")
    return full_str


def _get_hash_of_str(str_):
    h = hashlib.md5()
    h.update(str_.encode("utf-8"))
    return h.hexdigest()


def _get_applied_args(func, args, kwargs):
    # Get default args and kwargs
    signature = inspect.signature(func)
    applied_args = OrderedDict({k: v.default for k, v in signature.parameters.items()})

    # update call args into applied_args
    items = list(applied_args.items())
    for ix, arg in enumerate(args):
        applied_args[items[ix][0]] = arg
    for key, value in kwargs.items():
        applied_args[key] = value

    return applied_args



import csv
import datetime
import operator
import os
import re
import threading
import warnings

import chardet
import numpy as np
import pandas as pd
from Lutil._exceptions import DuplicateSettingWarning, SpeculationFailedError
from Lutil._logging import logger
from pandas.api.types import is_numeric_dtype, is_string_dtype

__all__ = ["DataReader", "AutoSaver"]


class DataReader(object):
    _instances = {}
    _instances_lock = threading.Lock()

    def __new__(cls, *args, **kwargs):
        if "_id" not in kwargs:
            _id = 0
        else:
            _id = kwargs["_id"]

        with DataReader._instances_lock:
            if _id in DataReader._instances:
                return DataReader._instances[_id]
            else:
                new_instance = object.__new__(cls)
                DataReader._instances[_id] = new_instance
                return new_instance

    def __init__(self, train_path=None, test_path=None, val_path=None, _id="default", read_func=None, **read_kwargs):
        assert read_func is None or callable(read_func)
        if hasattr(self, "_id"):
            self.__init_existed__(
                train_path=train_path,
                test_path=test_path,
                val_path=val_path,
                _id=_id,
                read_func=read_func,
                **read_kwargs,
            )
        else:
            self.__init_new__(
                train_path=train_path,
                test_path=test_path,
                val_path=val_path,
                _id=_id,
                read_func=read_func,
                **read_kwargs,
            )

    def __init_existed__(
        self, train_path=None, test_path=None, val_path=None, *, _id="default", read_func=None, **read_kwargs
    ):
        assert _id == self._id
        if train_path is not None:
            self.train_path = train_path
        if test_path is not None:
            self.test_path = test_path
        if val_path is not None:
            self.val_path = val_path

        if read_kwargs:
            if operator.eq(read_kwargs, self.__read_kwargs):
                logger.info(
                    f"Data reading configuration is already set for {self.__class__} object, it's unnecessary to set it again."
                )
            else:
                raise ValueError(
                    "Newly set data reading configuration is different from the cached value. If you do want this, please specify '_id=N' as a parameter."
                )

        if read_func:
            if read_func is self.__read_func:
                logger.info(
                    f"Data reading function is already set for {self.__class__} object, it's unnecessary to set it again."
                )
            else:
                raise ValueError(
                    "Newly set data reading function is different from the cached value. If you do want this, please specify '_id=N' as a parameter."
                )

    def __init_new__(
        self, train_path=None, test_path=None, val_path=None, *, _id="default", read_func=None, **read_kwargs
    ):
        try:
            if train_path is not None:
                assert os.path.exists(train_path)
            if test_path is not None:
                assert os.path.exists(test_path)
            if val_path is not None:
                assert os.path.exists(val_path)
        except Exception as e:
            raise ValueError("Some path is invalid or does not exist.", e)

        self._id = _id
        self.__read_kwargs = read_kwargs
        self.__read_func = read_func

        if self.__read_func is None:
            self.__read_func = pd.read_csv

        if train_path:
            self._train_path = train_path
        if test_path:
            self._test_path = test_path
        if val_path:
            self._val_path = val_path

    @property
    def train_path(self):
        if hasattr(self, "_train_path"):
            return self._train_path
        else:
            raise AttributeError(f"'{self.__class__}' object has no attribute 'train_path'.")

    @train_path.setter
    def train_path(self, value):
        if hasattr(self, "_train_path"):
            if self.train_path == value:
                logger.info(
                    f"'train_path' of {self.__class__} object is already set as {value}, it's unnecessary to set it again."
                )
            else:
                raise ValueError(
                    "Newly set train_path is different from the cached value. If you do want this, please specify '_id=N' as a parameter."
                )
        else:
            warnings.warn(DuplicateSettingWarning("train", self))
            self._train_path = value

    @property
    def test_path(self):
        if hasattr(self, "_test_path"):
            return self._test_path
        else:
            raise AttributeError(f"'{self.__class__}' object has no attribute 'test_path'.")

    @test_path.setter
    def test_path(self, value):
        if hasattr(self, "_test_path"):
            if self.test_path == value:
                logger.info(
                    f"'test_path' of {self.__class__} object is already set as {value}, it's unnecessary to set it again."
                )
            else:
                raise ValueError(
                    "Newly set test_path is different from the cached value. If you do want this, please specify '_id=N' as a parameter."
                )
        else:
            warnings.warn(DuplicateSettingWarning("test", self))
            self._test_path = value

    @property
    def val_path(self):
        if hasattr(self, "_val_path"):
            return self._val_path
        else:
            raise AttributeError(f"'{self.__class__}' object has no attribute 'val_path'.")

    @val_path.setter
    def val_path(self, value):
        if hasattr(self, "_val_path"):
            if self.val_path == value:
                logger.info(
                    f"'val_path' of {self.__class__} object is already set as {value}, it's unnecessary to set it again."
                )
            else:
                raise ValueError(
                    "Newly set val_path is different from the cached value. If you do want this, please specify '_id=N' as a parameter."
                )
        else:
            warnings.warn(DuplicateSettingWarning("val", self))
            self._val_path = value

    def train(self):
        return self.__read_func(self.train_path, **self.__read_kwargs)

    def test(self):
        return self.__read_func(self.test_path, **self.__read_kwargs)

    def val(self):
        return self.__read_func(self.val_path, **self.__read_kwargs)


class AutoSaver(object):
    def __init__(self, save_dir="", example_path=None, **default_kwargs):
        if example_path and default_kwargs:
            raise ValueError("You cannot set both 'example_path' and give other saving kwargs at the same time.")

        if save_dir and not os.path.exists(save_dir):
            os.mkdir(save_dir)

        self.save_dir = save_dir
        self.example_path = example_path
        self.default_kwargs = default_kwargs

    def __save_by_to_csv(self, X, filename):
        if self.example_path and self.__used_kwargs:
            raise ValueError("You cannot set both 'example_path' and give other saving kwargs at the same time.")

        if self.example_path:
            return self.__save_by_to_csv_speculating(X, filename)
        else:
            if not isinstance(X, (pd.DataFrame, pd.Series)):
                raise TypeError(
                    f"'X' must be a pd.DataFrame or pd.Series, rather than {X.__class__} if you do not provide an example csv file, or are using self-defined keyword parameters."
                )
            return X.to_csv(os.path.join(self.save_dir, filename), **self.__used_kwargs)

    def __speculate_ordered_index(self, s):
        s = s.copy()
        if is_string_dtype(s):
            return (True, s.iloc[0])

        step = len(s) // 100 + 1

        for i in range(0, len(s) - 1, step):
            if s.iloc[i + step] - s.iloc[i] != step:
                return (False,)
        return True, s.iloc[0]

    def __try_add_column(self, X):
        e = SpeculationFailedError("The number of columns of 'X' is smaller than that in the example file.")
        col_ix = self.__example_df.shape[1] - X.shape[1] - 1
        example_spec = self.__speculate_ordered_index(self.__example_df.iloc[:, col_ix])
        X = X.reset_index(level=0)
        if example_spec[0]:
            try:
                X.iloc[:, 0] = X.iloc[:, 0].astype(int)
            except:
                pass
            X_spec = self.__speculate_ordered_index(X.iloc[:, 0])
            if X_spec[0] and X_spec[1] == example_spec[1]:
                pass  # Index addition OK
            else:
                if example_spec[1] == 0:
                    X.iloc[:, 0] = np.arange(X.shape[0])
                elif example_spec[1] == 1:
                    X.iloc[:, 0] = np.arange(1, X.shape[0] + 1)
                else:
                    raise e
        elif X.shape[0] == self.__example_df.shape[0]:
            X.iloc[:, 0] = self.__example_df.iloc[:, 0]
        else:
            raise e

        return X

    def __get_example_df(self):
        with open(self.example_path, "rb") as f:
            buffer = f.read()
            enc = chardet.detect(buffer)["encoding"]

        with open(self.example_path, "r", encoding=enc) as f:
            df = pd.read_csv(f, header=None, nrows=1)

        has_header = is_string_dtype(df.iloc[0, :])

        with open(self.example_path, "r", encoding=enc) as f:
            sniffer = csv.Sniffer()
            content = f.read()
            try:
                dialect = sniffer.sniff(content)
                has_header = sniffer.has_header(content) or has_header

            except Exception:
                fixed_content = "\n".join(line + "," for line in content.split("\n"))
                dialect = sniffer.sniff(fixed_content)
                has_header = sniffer.has_header(fixed_content) or has_header

            dialect_kwargs = {
                "sep": dialect.delimiter,
                "line_terminator": dialect.lineterminator,
                "doublequote": dialect.doublequote,
                "quotechar": dialect.quotechar,
                "escapechar": dialect.escapechar,
                "quoting": dialect.quoting,
            }

            f.seek(0, 0)
            example_df = pd.read_csv(f, dialect=dialect, index_col=False, header=0 if has_header else None)
        self.__example_df = example_df
        self.__has_header = has_header
        self.__dialect_kwargs = dialect_kwargs

    def __save_by_to_csv_speculating(self, X, filename):
        if not isinstance(X, (pd.DataFrame, pd.Series, np.ndarray)):
            raise TypeError(
                f"'X' must be either a pd.DataFrame, pd.Series or np.ndarray, rather than {X.__class__} if you provide an example csv file."
            )

        X = pd.DataFrame(X)
        self.__get_example_df()

        while X.shape[1] > self.__example_df.shape[1]:
            # Drop columns of X
            if self.__speculate_ordered_index(X.iloc[:, 0])[0]:
                X = X.drop(columns=[X.columns.values[0]])
            else:
                raise SpeculationFailedError("The number of columns of 'X' is larger than that in the example file.")

        while X.shape[1] < self.__example_df.shape[1]:
            X = self.__try_add_column(X)

        for i in range(X.shape[1]):
            if self.__speculate_ordered_index(X.iloc[:, i])[0] and is_numeric_dtype(X.iloc[:, i]):
                X.iloc[:, i] = X.iloc[:, i].astype(int)
            else:
                break

        example_spec_res = self.__speculate_ordered_index(self.__example_df.iloc[:, 0])
        if example_spec_res[0] and is_numeric_dtype(example_spec_res[1]):
            X.iloc[:, 0] = np.arange(example_spec_res[1], example_spec_res[1] + X.shape[0])

        fullpath = os.path.join(self.save_dir, filename)

        if self.__has_header:
            header = [re.compile(r"Unnamed: \d+").sub("", i) for i in self.__example_df.columns.values]
            return X.to_csv(fullpath, header=header, index=False, **self.__dialect_kwargs)
        else:
            return X.to_csv(fullpath, header=False, index=False, **self.__dialect_kwargs)

    def save(self, X, filename=None, memo=None, **kwargs):
        if not filename:
            filename = datetime.datetime.now().strftime(r"%m%d-%H%M%S") + ".csv"
        self.__used_kwargs = {**self.default_kwargs, **kwargs}

        res = self.__save_by_to_csv(X, filename)

        if memo:
            with open(os.path.join(self.save_dir, "memo.txt"), "a+", encoding="utf-8") as f:
                f.write(f"{filename}: {str(memo)}" + "\n")

        return res



class SpeculationFailedError(Exception):
    def __init__(self, message):
        super().__init__("Failed to save by speculation, "+message)


class NotDecoratableError(Exception):
    def __init__(self, obj):
        super().__init__(
            f"{str(obj)} cannot be decorated by checkpoint, it may be a generator or a coroutine.")


class NotInlineCheckableError(Exception):
    def __init__(self, obj):
        super().__init__(
            f"{str(obj)} cannot be used for 'watch' parameter in InlineCheckpoint.")


class DuplicateSettingWarning(UserWarning):
    def __init__(self, name, obj):
        self.name = name
        self.obj = obj

    def __str__(self):
        return f"A {self.name} path is set after the first initialization of {self.obj.__class__}."


class ComplexParamsIdentifyWarning(UserWarning):
    def __init__(self, explain):
        self.explain = explain

    def __str__(self):
        return f"{self.explain}, it may cause mistake when detecting whether there is checkpoint for this call."


class InlineEnvironmentWarning(UserWarning):
    def __init__(self):
        super().__init__()

    def __str__(self):
        return '''InlineCheckpoint is only functional in '__main__' to auto-cache a variable, while you are trying to use it inside a function or method.
The 'produce' will not be cached and the computation will always happen.
Please refer to https://learning-utility.readthedocs.io/en/latest/ for alternative solutions.'''


class SkipWithBlock(Exception):
    pass



from setuptools import setup, find_packages

with open(r'README.md',"r",encoding="utf-8") as f:
    long_description = f.read()

setup(
    name='Lutil',
    version='0.1.10',
    author='Vopaaz',
    author_email='liyifan945@163.com',
    url='https://github.com/Vopaaz/learning-utility',
    description='Assist small-scale machine learning.',
    long_description = long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(),
    install_requires=["pandas", "chardet", "numpy", "joblib"],
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)"
    ],
)
